//
//   Version 1.0
//
//   Copyright 2016-2018 Lee Djavaherian
//
//   This program is free software: you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public License
//   along with this program.  If not, see <http://www.gnu.org/licenses/>.
///////////////////////////////////////////////////////////////////////////////////

#include<util/delay.h>
#include<avr/pgmspace.h>

#define T_DIT 1000
#define T_DAH (T_DIT * 3)
#define T_ENTER (T_DAH * 2)
#define T_HOLD (T_DAH * 4)
#define T_BIGPAUSE (T_HOLD * 2)
#define T_DEBOUNCE (T_DIT / 10)

//morse threshold 2-digit precision
#define PRECISION 10


//Threshold is a number based on inverted ratio T_DAH/T_DIT in 8-bit integer
//Normally this ratio is 3 per Morse standard (.333) or 3x
//But threshold should be a little less to allow normal Morse to be recognized
//Since floating point isn't used, instead, 22 is used, for dah 2.2x dit
//multiplied by the PRECISION (10) to gain the extra digit of precision
//Varying this threshold dynamically doesn't work - fixed cutoff works best
//The dit and dah lengths can be varied instead (used onratio below)
#define MORSE_THRESHOLD 20

#define KEYUP 1
#define KEYDOWN 0

#define DIT 0
#define DAH 1

//read_morse exit types
#define E_NONE 0
#define E_ENTER 1
#define E_HOLD 2
#define E_RESET 3


//global variables
volatile uint8_t morse_echo = 0; //disable immediate key echo on output
volatile uint8_t morse_replay = 0; //disable replay output of received input
volatile uint8_t pulseholder = 1;
volatile uint8_t lettercode = 0;
volatile uint16_t old_dit = T_DIT;
volatile uint16_t sound_dit = T_DIT; //value for sound only
volatile uint16_t old_dah = T_DAH;
void morse_vibrate (uint16_t length, uint8_t delay);
volatile uint8_t morse_to_letter (uint8_t pulses);
void letter_to_morse (uint8_t lettercode);
void insertion_sort(uint16_t A[], uint8_t positionA[]);
volatile uint8_t read_morse (void);

// timer 0 for internal routines 
// if you run at 1 Mhz, use prescaler of 64 (CS10 and CS11), then multiple T_DIT by 2
// since 64 is 1/16th of 1024
//TCCR1B |= (1 << CS10) | (1 << CS12); // Start timer 0 at Fcpu/1024
 // 0 bit is dit, 1 bit is dah, justified to the right
// with bit of 1 on far left for end marker

// the last 4 are not letters but special symbols
//struct bitfields don't work with with progmem, so shift/mask methods
//can be used to store and retrieve information in the unused 3-bits per byte in the array below
//an AND mask of 0b00011111 should be used to only retrieve the last 5 bits to allow
//the first 3 bits to be used for misc storage, freeing up around 11 bytes

const uint8_t morse[30] PROGMEM = {
                     0b00000101, // A 0  .-
                     0b00011000, // B 1  -...
                     0b00011010, // C 2  -.-.
                     0b00001100, // D 3  -..
                     0b00000010, // E 4  .
                     0b00010010, // F 5  ..-.
                     0b00001110, // G 6  --.
                     0b00010000, // H 7  ....
                     0b00000100, // I 8  ..
                     0b00010111, // J 9  .---
                     0b00001101, // K 10 -.-
                     0b00010100, // L 11 .-..
                     0b00000111, // M 12 --
                     0b00000110, // N 13 -.
                     0b00001111, // O 14 ---
                     0b00010110, // P 15 .--.
                     0b00011101, // Q 16 --.-
                     0b00001010, // R 17 .-.
                     0b00001000, // S 18 ...
                     0b00000011, // T 19 -
                     0b00001001, // U 20 ..-                   
                     0b00010001, // V 21 ...-
                     0b00001011, // W 22 .--
                     0b00011001, // X 23 -..-                     
                     0b00011011, // Y 24 -.--
                     0b00011100, // Z 25 --..

                     0b00011111,   // F1 ----
                     0b00011110,   // F2 ---.
                     0b00010101,   // F3 .-.-
                     0b00010011    // F4 ..--
};

/**************************************************************/
void morse_vibrate (uint16_t length, uint8_t delay)
{
   uint16_t mytime;
   TCCR1B |= (1 << CS10) | (1 << CS12); // Start timer 0 at Fcpu/1024
   mytime = TCNT1;
   while (TCNT1 < mytime + length) {
      
        if (morse_replay == 1) {MORSE_ON;}
        for (uint8_t delcnt = 0; delcnt < delay; delcnt++) {
          _delay_ms (1);
        }
        if (delay > 0) { //delay of 0 disables vibrate but keeps time delay intact
          if (morse_replay == 1) {MORSE_OFF;}
          for (uint8_t delcnt = 0; delcnt < delay; delcnt++) {
            _delay_ms (1);
          }
        } 
   }
   if (morse_replay == 1) {MORSE_OFF;}
}

/**************************************************************/

void letter_to_morse (uint8_t lettercode) {

TCCR1B |= (1 << CS10) | (1 << CS12); // Start timer 0 at Fcpu/1024
uint8_t letter;
uint16_t morsedelay; //used to match dit delays from TCNT1 to avoid _delay_ms function  

//converts integer hash of letter (A=0, Z=25) to blink morse LED
//halts program during this time which is ok temporarily

letter = pgm_read_byte(&(morse[lettercode])) & 0b00011111;
uint8_t bit = 0;

while ((letter & 128) != 128) {
  letter <<= 1; // if leftmost bit zero shift to left to find marker bit
  bit++;}

  letter <<= 1;  //shift off marker bit
  bit++;

  while (bit < 8) {  // now blink out the remaining bits
    if ((letter & 128) == 128) { 
     
       //DAH

       TCNT1 = 0;  //reset for sound
       morse_vibrate (T_DAH, VIBRATE_DELAY);
       TCNT1 = 0; //reset for delay
       morsedelay = TCNT1 + sound_dit;
       while (TCNT1 < morsedelay) {} //do nothing (add key interruption later)
    } else {

       // DIT

       TCNT1 = 0;  //reset for sound
       morse_vibrate (T_DIT, VIBRATE_DELAY);
       TCNT1 = 0; //reset for delay
       morsedelay = TCNT1 + sound_dit;
       while (TCNT1 < morsedelay) {} // do nothing (add key interruption later)
    } 

  letter <<= 1;
  bit++;
}
_delay_ms (600);
}

/****************************************************************/
volatile uint8_t morse_to_letter (uint8_t pulses) {

//converts morse string to integer hash of letter (A=0, Z=25)
//not cpu efficient, but saves lookup table space
//doesn't need to be fast

for (uint8_t letter = 0; letter < 30; letter++)
  if ((pgm_read_byte(&(morse[letter])) & 0b00011111) == pulses) return letter;



return 30; //if error for some reason 
}


/*****************************************************************/
//sorts main 16 bit array and keeps original position in 8-bit array
void insertion_sort(uint16_t main[], uint8_t position[])
{
   uint8_t morse_size = 4; uint8_t i;
   
   //first find size of array by looking for marker
   //set earlier in position array

   for (i = 1; i < 4; i++) {
     if (position[i] == 255) {morse_size = i;}
   }
   
   
   //now only sort array of that size
   uint16_t temp; uint8_t temp_position; uint8_t j;
   for (i = 1; i < morse_size; i++) {
     j = i;
     while (j > 0 && main[j-1] > main[j]) {
        temp = main[j]; temp_position = position[j];
        main[j] = main[j-1]; position[j] = position[j-1];
        main[j-1] = temp; position[j-1] = temp_position;
        j--;
     }
   }
   
     
   return;
}

/******************************************************************/

volatile uint8_t read_morse(void)
{
TCCR1B |= (1 << CS10) | (1 << CS12); // Start timer 0 at Fcpu/1024
  TCNT1=0; //key should be down, start the timer
  uint8_t pulseholder = 1;
  uint8_t exit_type = E_NONE;
  uint16_t onwave[4] = {65535, 65535, 65535, 65535}; //array to hold timer on values
  uint8_t position[4] = {0, 1, 2, 3};  //order of original positions

  uint8_t shh_dahflag = 0; //end DAH flag for SIMTHEO SHH only

  uint8_t morse_cnt=0;uint8_t key_state=KEYDOWN;if (morse_echo == 1) {MORSE_ON;}
  do {
     if (KEY_DOWN && key_state==KEYUP) { //KEY DOWN
        key_state=KEYDOWN;
        if (SIMTHEO_MODE == 1) {
           if (TCNT1 > T_DEBOUNCE)
           {     
//this doesn't actually run until after key is down, up, then down again
                 //first interrupt on key down is skipped
                 onwave[morse_cnt]=TCNT1; //store duration
                 morse_cnt++;
           }
        }
        TCNT1=0;

        if (morse_echo == 1) {MORSE_ON;} //LED on

     }
     else if (KEY_UP && key_state==KEYDOWN) //KEY UP
     {
        if (morse_echo == 1) {MORSE_OFF;} //LED off
        key_state=KEYUP;
        if (SIMTHEO_MODE == 0) {
           if (TCNT1 > T_DEBOUNCE)
           {           
              onwave[morse_cnt]=TCNT1; //store duration
              morse_cnt++;
           }
        } 
        TCNT1=0;
     }
     
     // flip led state to make sound while button down at 50 intervals
     if (key_state == KEYDOWN && (TCNT1 % 25 == 0)) {MORSE_FLIP;} 
     
     if (morse_cnt == 4 && key_state == KEYUP && TCNT1 > old_dah) {exit_type=E_ENTER;break;} //pause for a dah to prevent confusion
     if (key_state == KEYDOWN && TCNT1 > T_HOLD) {exit_type=E_HOLD;break;}
     if (SIMTHEO_MODE == 0 && key_state == KEYUP && TCNT1 > T_ENTER && morse_cnt > 0) {exit_type=E_ENTER;break;}
     if (SIMTHEO_MODE == 1 && key_state == KEYUP && TCNT1 > T_BIGPAUSE) {onwave[morse_cnt]=TCNT1;exit_type=E_ENTER;break;}
  

  } while(1);  //loops until morse entered (main loop is stopped)

  //now morse wave has been captured to array (on times only for SIMTHEO, off times for SIMTHEO SHH)

  if (exit_type==E_ENTER) {
    
    if (SIMTHEO_MODE == 1) {
       //CHECK FOR TRAILING DAH
       //make 4th element a dah if it exited early due to 5 pulses (8 possibilies: JQVXY, F1, F3, F4), otherwise, find out if last element exceeded the hold pause length to set DAH flag for SIMTHEO SHH only     
       if ((morse_cnt == 4) || (morse_cnt > 0 && onwave[morse_cnt-1] > T_HOLD)) {shh_dahflag = 1;morse_cnt--;}   
    }


    //now delimit smaller onwave array by marking endpoint in position array
    if (morse_cnt < 4) {position[morse_cnt] = 255;}    
    
    //now sort arrays
    insertion_sort(onwave, position);  
     
    //the times 10 on the numerator scales the result for more precision for the threshold
    //however if timer gets too large, it will overflow the 16 bit integer,
    //but this should not be possible due to hold activating

    //ratio of longest to shortest, scaling for integers instead of float - use divide instead of multiply to avoid integer overflow which occurs otherwise
    uint16_t onratio = (onwave[morse_cnt-1]/(onwave[0] / PRECISION)); 
    
    uint16_t ontime_a; uint16_t ontime_b;
    
    if (onratio > MORSE_THRESHOLD) {
       /////////////// mixed dits and dahs - now find gap between them ////////////

       //Since mode is mixed, reset sames to not adjusted
       //This allows same dit/dah length adjustment to occur only once after mixed       

       uint16_t gap = 0; uint8_t splitpoint = 0;
       for (uint8_t cnt=1;cnt < morse_cnt;cnt++) {
          if (onwave[cnt] > onwave[cnt-1] && onwave[cnt] - onwave[cnt-1] > gap) {gap = onwave[cnt] - onwave[cnt-1];splitpoint=cnt;}
          else if (onwave[cnt] < onwave[cnt-1] && onwave[cnt-1] - onwave[cnt] > gap) {gap = onwave[cnt-1] - onwave[cnt];splitpoint=cnt;}          
       } 
       
       
       
       //first do dahs
       //median works well since avg can create rounding problems for integers
       //and using median also eliminates outliers
       //set to actual if 1, get avg (if 2), otherwise get median (if 3)
       if (morse_cnt - splitpoint == 1) {
          ontime_a = onwave[morse_cnt - 1]; //only one for dah
       } 
       //average of 2 dahs
       //note that avg function is higher - (higher - lower)/2 to avoid integer overflow  
       else if (morse_cnt - splitpoint == 2) {
           ontime_a = onwave[morse_cnt - 1] - ((onwave[morse_cnt - 1] - onwave[splitpoint]) / 2);           
        }
        else {     
           ontime_a = onwave[2]; //median for dah if 3
        }
        
       // now do dits       
       if (splitpoint == 1) {
          ontime_b = onwave[0]; //only one for dit
       } 
       //average of 2 dits
       //note that avg function is higher - (higher - lower)/2 to avoid integer overflow  
       else if (splitpoint == 2) {
           ontime_b = onwave[1] - ((onwave[1] - onwave[0]) / 2);           
        }
        else {     
           ontime_b = onwave[1]; //median for dit if 3
        }

          old_dit = ontime_b;
          old_dah = ontime_a;

       //now overwrite relevant section of onwave array to mark discrete dits and dahs
       uint8_t cnt;       
       for (cnt=0;cnt < splitpoint;cnt++) {
          onwave[position[cnt]] = DIT;
       }
       for (cnt=splitpoint;cnt < morse_cnt;cnt++) {
          onwave[position[cnt]] = DAH;
        }
        sound_dit = old_dit;
    }     
    else {
       //////////// all dits or all dahs - determine which one //////////
       
       
       if (morse_cnt == 4) {
          //average 2 center ones (combo avg/mean)
          ontime_a = onwave[2] - ((onwave[2] - onwave[1]) / 2);
       }     
       else if (morse_cnt == 3) { 
          ontime_a = onwave[1]; //median of 3
       } 
       //average of only 2
       else if (morse_cnt == 2) {
           ontime_a = onwave[1] - ((onwave[1] - onwave[0]) / 2); 
        }
        else {
           ontime_a = onwave[0]; //only one
        }

       //now find which one it is closest to using old_dit and old_dah
       //that were calculated from the last mixed dit/dah letter
       //can't use abs function, so using comparisons
       //Allow a dit/dah length adjustment only once after mixed by monitoring same_dit/same_dah flags       

       uint8_t pulsetype;
       if (ontime_a <= old_dit) { //shorter than dit
               pulsetype = DIT;
               sound_dit = ontime_a;

       }
       else if (ontime_a >= old_dah) { //longer than dah
               pulsetype = DAH;
               sound_dit = ontime_a/3;
       }
       //in-between dit and dah - find closest
       else if ((ontime_a - old_dit) < (old_dah - ontime_a))  { //closer to dit
               pulsetype = DIT; 
               sound_dit = ontime_a;            
       }
       else //closer to dah
       { pulsetype = DAH;
         sound_dit = ontime_a/3;
        }

       //now fill array with that same type
       for (uint8_t cnt=0;cnt < morse_cnt;cnt++) {
          onwave[cnt] = pulsetype;
       }   

    }
    
    
    // now convert array to pulseholder
    for (uint8_t cnt=0;cnt < morse_cnt;cnt++) 
    {
       if (onwave[cnt] == DIT)
       {    
             pulseholder <<= 1; // left shift to clear right spot, make zero
       }       
       else if (onwave[cnt] == DAH) //maybe remove this check since implied
       {
             pulseholder <<= 1; // left shift to clear spot
             pulseholder |= 1; // set rightmost bit to 1
       }
       
    }    

    if (SIMTHEO_MODE == 1) {
       pulseholder <<= 1; //add final dit
       if (shh_dahflag == 1) {pulseholder |= 1; shh_dahflag = 0;} // change ending dit to dah instead, since dahflag was entered
    }   
    
      //now replay the output what you just received
      
         uint8_t lettercode = morse_to_letter(pulseholder);
         if (lettercode == 26 && SIMTHEO_MODE == 0) {lettercode = 7;} //fix ---- 
         if (lettercode >=0 && lettercode <= 25) {
                letter_to_morse (lettercode);        
                //call main routine - later
         }

else if (lettercode > 25 && lettercode < 30 )
           {
           TCNT1 = 0;  //reset for sound   
           morse_vibrate (T_DAH, 255); 
         }  

      _delay_ms(1000); //add 1 second delay (in case two systems are running in parallel)
      return lettercode;
    
  }          
  else if (exit_type==E_HOLD) // perform hold function
  { 
    old_dit = T_DIT;
    sound_dit = T_DIT;
    old_dah = T_DAH;
    return 31;
  }
    else if (exit_type==E_RESET) // perform reset function
  {     
    return 32;
  }
//sleep chip - later

} 


